class CTVersionManager():

  def __init__(self):
    self.versions = {}

  def clear(self):
    self.versions = {}

  def add(self, name, value):
    self.versions[name] = value

  def get(self, name):
    if name in self.versions:
      return self.versions[name]
    else:
      return ""
    
ct_version_manager = CTVersionManager()




from .api_base_model import ApiBaseModel
from .code import Code

class Address(ApiBaseModel):
  addressId: str
  text: str
  line: str
  city: str
  district: str
  state: str
  postalCode: str
  country: Code



from enum import Enum
from typing import List, Dict, Union

from .api_base_model import ApiBaseModel
from .timing import Timing

class ScheduledInstanceType(Enum):
  ACTIVITY = 1
  DECISION = 2

class ScheduledInstance(ApiBaseModel):
    scheduledInstanceId: str
    scheduledInstanceType: ScheduledInstanceType
    scheduleSequenceNumber: int
    scheduleTimelineExitId: Union[str, None] = None
    scheduledInstanceEncounterId: Union[str, None] = None
    scheduledInstanceTimings: List[Timing] = []
    scheduledInstanceTimelineId: Union[str, None] = None

class ScheduledActivityInstance(ScheduledInstance):
    activityIds: List[str] = []

class ScheduledDecisionInstance(ScheduledInstance):
    conditionAssignments: List[List] = []


import pandas as pd
from usdm_excel.id_manager import id_manager
from usdm_excel.cdisc_ct import CDISCCT
from usdm_model.code import Code
from usdm_excel.ct_version_manager import ct_version_manager
from usdm_excel.errors.errors import error_manager
from usdm_excel.logger import package_logger

class BaseSheet():

  def __init__(self, file_path, sheet_name, header=0):
    self.file_path = file_path
    self.sheet_name = sheet_name
    self.sheet = pd.read_excel(open(file_path, 'rb'), sheet_name=sheet_name, header=header)
    self._general_info("Processed sheet %s" % (sheet_name))

  def cell_empty(self, row_index, col_index):
    return pd.isnull(self.sheet.iloc[row_index, col_index])  

  def read_cell_by_name(self, row_index, field_name):
    try:
      col_index = self.sheet.columns.get_loc(field_name)
      return self.read_cell(row_index, col_index)
    except Exception as e:
      col_index = -2
      self._error(row_index, col_index, "Error (%s) reading cell" % (e))
      return ""

  def read_cell(self, row_index, col_index):
    try:
      if pd.isnull(self.sheet.iloc[row_index, col_index]):
        return ""
      else:
        return str(self.sheet.iloc[row_index, col_index]).strip()
    except Exception as e:
      self._error(row_index, col_index, "Error (%s) reading cell" % (e))
      return ""

  def read_cell_empty_legacy(self, row_index, col_index):
    if self.cell_empty(row_index, col_index):
      return "", True
    else:
      value = self.read_cell_empty(row_index, col_index, '-')
      if value == "":
        return "", True
      else:
        return value, False

  def read_cell_empty(self, row_index, col_index, empty_character):
    value = self.read_cell(row_index, col_index)
    value = "" if value == empty_character else value
    return value

  def read_cell_multiple(self, rindex, cindex):
    results = []
    value = self.read_cell(rindex, cindex)
    if value.strip() == '':
      return results
    for part in value.split(","):
      results.append(part.strip())
    return results

  def read_cell_with_previous(self, row_index, col_index, first_col_index):
    try:
      i = col_index
      while i >= first_col_index:
        if pd.isnull(self.sheet.iloc[row_index, i]):
          i -= 1
        else:
          return self.sheet.iloc[row_index, i].strip()
      self._error(row_index, col_index, "Blank cell error")
      return ""
    except Exception as e:
      self._error(row_index, col_index, "Error (%s) reading cell row '%s', field '%s'" % (e, row_index, col_index))
      return ""

  def read_boolean_cell_by_name(self, row_index, field_name):
    value = self.read_cell_by_name(row_index, field_name)
    if value.strip().upper() in ['Y', 'YES', 'TRUE', '1']:
      return True
    return False

  # Want to kill this method
  def set_cdisc_code(self, value):
    if value.strip() == "":
      return None
    parts = value.split("=")
    try:
      return CDISCCT().code(code=parts[0].strip(), decode=parts[1].strip())
    except Exception as e:
      self._error(0, 0, "CDISC code error '%s'" % (e))
      return None

  def read_other_code_cell_by_name(self, row_index, field_name):
    col_index = self.sheet.columns.get_loc(field_name)
    return self.read_other_code_cell(row_index, col_index)

  def read_other_code_cell(self, row_index, col_index):
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      return None
    return self._decode_other_code(value, row_index, col_index)

  def read_other_code_cell_multiple_by_name(self, row_index, field_name):
    col_index = self.sheet.columns.get_loc(field_name)
    return self.read_other_code_cell_mutiple(row_index, col_index)

  def read_other_code_cell_mutiple(self, row_index, col_index):
    value = self.read_cell(row_index, col_index)
    result = []
    if value.strip() == '':
      return result
    items = value.split(",")
    for item in items:
      code = self._decode_other_code(item.strip(), row_index, col_index)
      if not code == None:
        result.append(code)
    return result

  def read_cdisc_klass_attribute_cell_by_name(self, klass, attribute, row_index, field_name):
    col_index = self.sheet.columns.get_loc(field_name)
    return self.read_cdisc_klass_attribute_cell(klass, attribute, row_index, col_index)

  def read_cdisc_klass_attribute_cell(self, klass, attribute, row_index, col_index):
    code = None
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where CDISC CT value expected.")
    else:
      code = CDISCCT().code_for_attribute(klass, attribute, value)
      if code is None:
        self._error(row_index, col_index, f"CDISC CT not found for value '{value}'.")
    return code

  def read_cdisc_klass_attribute_cell_multiple_by_name(self, klass, attribute, row_index, field_name):
    col_index = self.sheet.columns.get_loc(field_name)
    return self.read_cdisc_klass_attribute_cell_multiple(klass, attribute, row_index, col_index)

  def read_cdisc_klass_attribute_cell_multiple(self, klass, attribute, row_index, col_index):
    result = []
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where multiple CDISC CT values expected.")
      return result
    items = value.split(",")
    for item in items:
      code = CDISCCT().code_for_attribute(klass, attribute, item.strip())
      if code is not None:
        result.append(code)
      else:
        self._error(row_index, col_index, f"CDISC CT not found for value '{item.strip()}'.")
    return result

  def double_link(self, items, id, prev, next):
    for idx, item in enumerate(items):
      if idx == 0:
        setattr(item, prev, None)
      else:
        the_id = getattr(items[idx-1], id)
        setattr(item, prev, the_id)
      if idx == len(items)-1:  
        setattr(item, next, None)
      else:
        the_id = getattr(items[idx+1], id)
        setattr(item, next, the_id)

  def _decode_other_code(self, value, row_index, col_index):
    if value.strip() == "":
      return None
    outer_parts = value.split(":")
    if len(outer_parts) == 2:
      system = outer_parts[0].strip()
      inner_parts = outer_parts[1].strip().split("=")
      if len(inner_parts) == 2:
        version = ct_version_manager.get(system)
        return Code(codeId=id_manager.build_id(Code), code=inner_parts[0].strip(), codeSystem=system, codeSystemVersion=version, decode=inner_parts[1].strip())
      else:
        self._error(row_index, col_index, "Failed to decode code data '%s', no '=' detected" % (value))
    else:
      self._error(row_index, col_index, "Failed to decode code data '%s', no ':' detected" % (value))
    return None

  def _info(self, row, column, message):
     package_logger.info(self._format(row + 1, column + 1, message))
     
  def _general_info(self, message):
     package_logger.info(self._format(None, None, message))
     
  def _error(self, row, column, message):
    error_manager.add(self.sheet_name, row + 1, column + 1, message)

  def _general_error(self, message):
    error_manager.add(self.sheet_name, None, None, message)

  def _warning(self, row, column, message):
    error_manager.add(self.sheet_name, row + 1, column + 1, message, error_manager.WARNING)

  def _traceback(self, message):
    package_logger.debug(message)

  def _format(self, row, column, message):
    if self.sheet_name == None:
      return f"{self.message}"
    elif row == None:
      return f"In sheet {self.sheet_name}: {message}"
    else:
      return f"In sheet {self.sheet_name} at [{row},{column}]: {message}"



from usdm_excel.id_manager import id_manager
from usdm_model.code import Code
import requests
import os
import yaml
from usdm_excel.cdisc_ct_library import cdisc_ct_library

class CDISCCT():

  API_ROOT = 'https://api.library.cdisc.org/api'  

  def code(self, code, decode):
    return Code(codeId=id_manager.build_id(Code), code=code, codeSystem=cdisc_ct_library.system, codeSystemVersion=cdisc_ct_library.version, decode=decode)
 
  def code_for_attribute(self, klass, attribute, value):
    item = cdisc_ct_library.klass_and_attribute(klass, attribute, value)
    if item == None:
      return None
    return Code(codeId=id_manager.build_id(Code), code=item['conceptId'], codeSystem=cdisc_ct_library.system, codeSystemVersion=cdisc_ct_library.version, decode=item['preferredTerm'])



import stringcase
import json

class NodesAndEdges():

  def __init__(self, study):
    self.study = study
    self.nodes = []
    self.edges = []
    self.add_edges = []
    self.node_index = 1
    self.edge_index = 1
    self.id_node_index_map = {}
    self.edge_attributes = [
      'encounterIds',
      'timepointActivityIds',
      'timepointEncounterId',
      'bcSurrogateIds',
      'bcCategoryIds',
      'biomedicalConceptIds',
      'biomedicalConceptSurrogateId',
      'relativeFromScheduledInstanceId',
      'relativeToScheduledInstanceId',
      'scheduledInstanceEncounterId',
      'activityIds',
      'scheduledDecisionInstanceId',
      'treatment',
      'variableOfInterest',
      'conditionAssignments',
      'activityTimelineId'
    ]
    self.fix_id_name = {
      'scheduledActivityInstanceId': 'scheduledInstanceId',
      'scheduledDecisionInstanceId': 'scheduledInstanceId',
      'biomedicalConceptSurrogateId': 'bcSurrogateId',
      'biomedicalConceptPropertyId': 'bcPropertyId'
    }
  
  def nodes_and_edges(self):
    node = json.loads(self.study.to_json_with_type())
    self._process_node(node)
    for edge in self.add_edges:
      if edge['end'] in self.id_node_index_map:
        edge['id'] = self.edge_index
        edge['end'] = self.id_node_index_map[edge['end']]
        self.edges.append(edge)
        self.edge_index += 1
      else:
        print("***** %s -edge-> %s [%s] *****" % (edge['start'], edge['end'], edge['properties']))
    return self.nodes, self.edges
  
  def _process_node(self, node):
    if type(node) == list:
      result = []
      for item in node:
        indexes = self._process_node(item)
        result = result + indexes
      return result
    elif type(node) == dict:
      if node == {}:
        return []
      properties = {}
      id_field, klass = self._get_id_field_and_klass(node)
      if node[id_field] in self.id_node_index_map:
        return [self.id_node_index_map[node[id_field]]]
      this_node_index = self.node_index
      self.node_index += 1
      for key, value in node.items():
        if key in self.edge_attributes:
          if key == "conditionAssignments":
            # Special case, array of arrays of condition and link id
            for item in value:
              self.add_edges.append( { 'start': this_node_index, 'end': item[1], 'properties': { 'label': key, 'type': 'Condition' }})
          elif type(value) == list:
            for item in value:
              self.add_edges.append( { 'start': this_node_index, 'end': item, 'properties': { 'label': key, 'type': 'List' }})
          else:
            if not value == "":
              self.add_edges.append( { 'start': this_node_index, 'end': value, 'properties': { 'label': key, 'type': 'Other' }})
            # else:
            #   print("Key %s, value >%s<" % (key, value))
        else:
          indexes = self._process_node(value)
          if indexes == []:
            properties[key] = value
          else:
            for index in indexes:
              self.edges.append( {'id': self.edge_index, 'start': this_node_index, 'end': index, 'properties': {'label': key}})
              self.edge_index += 1
      if klass == "Study":
        node[id_field] = "Study"
      properties['label'] = node[id_field]
      self.nodes.append({ 'id': this_node_index, 'properties': properties })
      self.id_node_index_map[properties[id_field]] = this_node_index
      return [this_node_index]
    else:
      return []
    
  def _get_id_field_and_klass(self, node):
    klass = node['_type']
    id_name = "%s%s" % (stringcase.camelcase(klass), "Id")
    if id_name in self.fix_id_name:
      id_name = self.fix_id_name[id_name]
    return id_name, klass




from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.study_epoch import StudyEpoch
from usdm_model.study_arm import StudyArm
from usdm_model.study_cell import StudyCell
from usdm_model.study_design import StudyDesign
from usdm_excel.alias import Alias
import pandas as pd
import traceback

class StudyDesignSheet(BaseSheet):

  NAME_ROW = 0
  DESCRIPTION_ROW = 1
  TA_ROW = 2
  RATIONALE_ROW = 3
  BLINDING_ROW = 4
  INTENT_ROW = 5
  TYPES_ROW = 6
  INT_ROW = 7
  MAIN_TIMELINE_ROW = 8
  OTHER_TIMELINES_ROW = 9

  EPOCH_ARMS_START_ROW = 11
  
  PARAMS_DATA_COL = 1

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='studyDesign', header=None))
      super().__init__(file_path=file_path, sheet_name='studyDesign', header=None)
      self.name = "TEST"
      self.description = "An Microsoft Excel test study design"
      self.epochs = []
      self.arms = []
      self.cells = []
      self.study_designs = []
      self.therapeutic_areas = []
      self.rationale = ""
      self.blinding = None
      self.trial_intents = []
      self.trial_types = []
      self.intervention_model = None
      self.main_timeline = None
      self.other_timelines = []
      self.process_sheet()
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def process_sheet(self):
    for rindex, row in self.sheet.iterrows():
      if rindex == self.NAME_ROW:
        self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.DESCRIPTION_ROW:
        self.description = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.TA_ROW:
        self.therapeutic_areas = self.read_other_code_cell_mutiple(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.RATIONALE_ROW:
        self.rationale = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.BLINDING_ROW:
        self.blinding = Alias().code(self.read_cdisc_klass_attribute_cell('StudyDesign', 'studyDesignBlindingScheme',rindex, self.PARAMS_DATA_COL), [])
      elif rindex == self.INTENT_ROW:
        self.trial_intents = self.read_cdisc_klass_attribute_cell_multiple('StudyDesign', 'trialIntentType', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.TYPES_ROW:
        self.trial_types = self.read_cdisc_klass_attribute_cell_multiple('StudyDesign', 'trialType', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.INT_ROW:
        self.intervention_model = self.read_cdisc_klass_attribute_cell('StudyDesign', 'interventionModel', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.MAIN_TIMELINE_ROW:
        self.main_timeline = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.OTHER_TIMELINES_ROW:
        self.other_timelines = self.read_cell_multiple(rindex, self.PARAMS_DATA_COL)
      else:
        pass

    for rindex, row in self.sheet.iterrows():
      if rindex >= self.EPOCH_ARMS_START_ROW:
        for cindex in range(0, len(self.sheet.columns)):
          cell = self.read_cell(rindex, cindex)
          if rindex == self.EPOCH_ARMS_START_ROW:
            if cindex != 0:
              epoch = self._add_epoch(cell, cell)
              self.epochs.append(epoch)
          else:
            if cindex == 0:
              self.arms.append(self._add_arm(cell, cell))
            else:
              elements = []
              element_names = self.read_cell_multiple(rindex, cindex)
              for name in element_names:
                elements.append(cross_references.get(name))
              self.cells.append(self._add_cell(arm=self.arms[-1], epoch=self.epochs[cindex-1], elements=elements))
    
    self.double_link(self.epochs, 'studyEpochId', 'previousStudyEpochId', 'nextStudyEpochId')

    study_design = self._add_design(
      name=self.name,
      description=self.description,
      cells=self.cells, 
      intent_types=self.trial_intents, 
      trial_types=self.trial_types, 
      intervention_model=self.intervention_model,
      rationale=self.rationale, 
      blinding=self.blinding, 
      therapeutic_areas=self.therapeutic_areas
    )
    self.study_designs.append(study_design)

  def _add_arm(self, name, description):
    # TODO read the arm origin and type
    arm_origin = self.set_cdisc_code("C188866=Data Generated Within Study")
    arm_type = self.set_cdisc_code("C174266=Experimental Arm")
    return StudyArm(
      studyArmId=id_manager.build_id(StudyArm), 
      studyArmName=name,
      studyArmDescription=description,
      studyArmType=arm_type,
      studyArmDataOriginDescription="",
      studyArmDataOriginType=arm_origin
    )

  def _add_epoch(self, name, description):
    # TODO read the epoch type
    epoch_type = self.set_cdisc_code("C165873=OBSERVATION")
    return StudyEpoch(
      studyEpochId=id_manager.build_id(StudyEpoch), 
      studyEpochName=name, 
      studyEpochDescription=description,
      studyEpochType=epoch_type
    )
  
  def _add_cell(self, arm, epoch, elements):
    return StudyCell(
      studyCellId=id_manager.build_id(StudyCell), 
      studyArm=arm,
      studyEpoch=epoch,
      studyElements=elements
    )

  def _add_design(self, name, description, cells, intent_types, trial_types, intervention_model, rationale, blinding, therapeutic_areas):
    return StudyDesign(
      studyDesignId=id_manager.build_id(StudyDesign), 
      studyDesignName=name,
      studyDesignDescription=description,
      trialIntentTypes=intent_types,
      trialType=trial_types,
      interventionModel=intervention_model,
      studyCells=cells,
      therapeuticAreas=therapeutic_areas,
      studyDesignRationale=rationale,
      studyDesignBlindingScheme=blinding
    )



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
import pandas as pd

class Encounters():
  
  def __init__(self, parent):
    self.parent = parent
    self._epoch_map = {}
    for col_index in range(self.parent.sheet.shape[1]):
      if col_index >= SoAColumnRows.FIRST_VISIT_COL:
        self.xref, is_null = self.parent.read_cell_empty_legacy(SoAColumnRows.VISIT_LABEL_ROW, col_index)
        self.epoch = self.parent.read_cell_with_previous(SoAColumnRows.EPOCH_ROW, col_index, SoAColumnRows.FIRST_VISIT_COL)
        if not self.xref == "":
          encounter = cross_references.get(self.xref)
          if self.epoch not in self._epoch_map:
            self._epoch_map[self.epoch] = []
          self._epoch_map[self.epoch].append(encounter.encounterId)

  def epoch_encounter_map(self, epoch):
    return self._epoch_map[epoch]



from src.usdm_excel.errors.errors import Errors

def test_create():
  errors = Errors()
  assert errors.items == []

def test_add():
  errors = Errors()
  assert len(errors.items) == 0
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert len(errors.items) == 1
  assert errors.items[0].message == "XXXXX"
    
def test_count():
  errors = Errors()
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert errors.count() == 2

def test_clear():
  errors = Errors()
  assert len(errors.items) == 0
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert len(errors.items) == 1
  errors.clear()
  assert len(errors.items) == 0

def test_dumo():
  errors = Errors()
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX1")
  errors.add(sheet="My Sheet", row=2, column=100, message="XXXXX2")
  assert (errors.dump()) == [
    {'sheet': 'My Sheet', 'row': 1, 'column': 99, 'message': 'XXXXX1', 'level': 40},
    {'sheet': 'My Sheet', 'row': 2, 'column': 100, 'message': 'XXXXX2', 'level': 40}
  ]



from typing import List, Union
from .api_base_model import ApiBaseModel
from .code import Code
from uuid import UUID

class Indication(ApiBaseModel):
  indicationId: str
  indicationDescription: str
  codes: List[Code] = []



from .api_base_model import ApiBaseModel
from typing import List
from .code import Code

class StudyDesignPopulation(ApiBaseModel):
  studyDesignPopulationId: str
  populationDescription: str
  plannedNumberOfParticipants: int
  plannedMaximumAgeOfParticipants: str  
  plannedMinimumAgeOfParticipants: str
  plannedSexOfParticipants: List[Code] = []



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_identifiers_sheet.study_identifiers_sheet import StudyIdentifiersSheet
from usdm_excel.study_design_sheet.study_design_sheet import StudyDesignSheet
from usdm_excel.study_soa_sheet.study_soa_sheet import StudySoASheet
from usdm_excel.study_design_ii_sheet.study_design_ii_sheet import StudyDesignIISheet
from usdm_excel.study_design_population_sheet.study_design_population_sheet import StudyDesignPopulationSheet
from usdm_excel.study_design_objective_endpoint_sheet.study_design_objective_endpoint_sheet import StudyDesignObjectiveEndpointSheet
from usdm_excel.study_design_estimands_sheet.study_design_estimands_sheet import StudyDesignEstimandsSheet
from usdm_excel.study_design_procedure_sheet.study_design_procedure_sheet import StudyDesignProcedureSheet
from usdm_excel.study_design_encounter_sheet.study_design_encounter_sheet import StudyDesignEncounterSheet
from usdm_excel.study_design_element_sheet.study_design_element_sheet import StudyDesignElementSheet
from usdm_excel.alias import Alias
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.study import Study
from usdm_model.study_protocol_version import StudyProtocolVersion
import traceback
import pandas as pd
import datetime

class StudySheet(BaseSheet):

  TITLE_ROW = 0
  VERSION_ROW = 1
  TYPE_ROW = 2
  PHASE_ROW = 3
  ACRONYM_ROW = 4
  RATIONALE_ROW = 5
  TA_ROW = 6

  PROTOCOL_HEADER_ROW = 8
  PROTOCOL_DATA_ROW = 9
  
  PARAMS_DATA_COL = 1

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='study', header=None))
      super().__init__(file_path=file_path, sheet_name='study', header=None)
      self.phase = None
      self.version = None
      self.type = None
      self.title = None
      self.acronym = None
      self.rationale = None
      self.study = None
      self.protocols = []
      self.therapeutic_areas = []
      self.timelines = {}
      self._process_sheet()
      self.study_identifiers = StudyIdentifiersSheet(file_path)
      self.procedures = StudyDesignProcedureSheet(file_path)
      self.encounters = StudyDesignEncounterSheet(file_path)
      self.elements = StudyDesignElementSheet(file_path)
      self.study_design = StudyDesignSheet(file_path)
      for timeline in self.study_design.other_timelines:
        tl = StudySoASheet(file_path, timeline)
        self.timelines[timeline] = tl
        cross_references.add(timeline, tl.timeline.scheduleTimelineId)
      self.soa = StudySoASheet(file_path, self.study_design.main_timeline)
      self.ii = StudyDesignIISheet(file_path)
      self.study_populations = StudyDesignPopulationSheet(file_path)
      self.oe = StudyDesignObjectiveEndpointSheet(file_path)
      self.estimands = StudyDesignEstimandsSheet(file_path)

      for epoch in self.study_design.epochs:
        epoch.encounterIds = self.soa.epoch_encounter_map(epoch.studyEpochName)

      study_design = self.study_design.study_designs[0]
      study_design.studyScheduleTimelines.append(self.soa.timeline)
      study_design.encounters = self.encounters.items
      study_design.activities = self.soa.activities
      study_design.biomedicalConcepts = self.soa.biomedical_concepts
      study_design.bcSurrogates = self.soa.biomedical_concept_surrogates
      for key,tl in self.timelines.items():
        study_design.studyScheduleTimelines.append(tl.timeline)
        study_design.activities += tl.activities
        study_design.biomedicalConcepts += tl.biomedical_concepts
        study_design.bcSurrogates += tl.biomedical_concept_surrogates
      study_design.studyIndications = self.ii.indications
      study_design.studyInvestigationalInterventions = self.ii.interventions
      study_design.studyPopulations = self.study_populations.populations
      study_design.studyObjectives = self.oe.objectives
      study_design.studyEstimands = self.estimands.estimands

      self.study = Study(
        studyId=None, # No Id, will be allocated a UUID
        studyTitle=self.title,
        studyVersion=self.version,
        studyType=self.type,
        studyPhase=self.phase,
        businessTherapeuticAreas=self.therapeutic_areas,
        studyRationale=self.rationale,
        studyAcronym=self.acronym,
        studyIdentifiers=self.study_identifiers.identifiers,
        studyProtocolVersions=self.protocols,
        studyDesigns=self.study_design.study_designs
      )
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def study_sponsor(self):
    return self.cdisc_code(code="C93453", decode="Study Registry")

  def study_regulatory(self):
    return self.cdisc_code(code="C93453", decode="Study Registry")

  def the_study(self):
    return self.study
  
  def _process_sheet(self):
    fields = [ 'briefTitle', 'officialTitle', 'publicTitle', 'scientificTitle', 'protocolVersion', 'protocolAmendment', 'protocolEffectiveDate', 'protocolStatus' ]    
    for rindex, row in self.sheet.iterrows():
      if rindex == self.TITLE_ROW:
        self.title = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROW:
        self.version = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.TYPE_ROW:
        self.type = self.read_cdisc_klass_attribute_cell('Study', 'studyType', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.PHASE_ROW:
        phase = self.read_cdisc_klass_attribute_cell('Study', 'studyPhase', rindex, self.PARAMS_DATA_COL)
        self.phase = Alias().code(phase, [])
      elif rindex == self.ACRONYM_ROW:
        self.acronym = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.RATIONALE_ROW:
        self.rationale = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.TA_ROW:
        self.therapeutic_areas = self.read_other_code_cell_mutiple(rindex, self.PARAMS_DATA_COL)
        #print("8", self.therapeutic_areas)
      elif rindex >= self.PROTOCOL_DATA_ROW:
        record = {}
        for cindex in range(0, len(self.sheet.columns)):
          field = fields[cindex]
          if field == 'protocolStatus':
            record[field] = self.read_cdisc_klass_attribute_cell('StudyProtocolVersion', 'protocolStatus', rindex, cindex) 
          elif field == 'protocolEffectiveDate':
            cell = self.read_cell(rindex, cindex)
            record[field] = datetime.datetime.strptime(cell, '%Y-%m-%d %H:%M:%S')
          else:
            cell = self.read_cell(rindex, cindex)
            record[field] = cell
        record['studyProtocolVersionId'] = id_manager.build_id(StudyProtocolVersion)
        spv = StudyProtocolVersion(**record)
        self.protocols.append(spv)
  


from typing import List
from .api_base_model import ApiBaseModel

class BiomedicalConceptCategory(ApiBaseModel):
  biomedicalConceptCategoryId: str
  bcCategoryParentIds: List[str] = []
  bcCategoryChildrenIds: List[str] = []
  bcCategoryName: str
  bcCategoryDescription: str
  bcCategoryMemberIds: List[str] = []



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.study_element import StudyElement
from usdm_model.transition_rule import TransitionRule

class StudyDesignElementSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='studyDesignElements'))
      super().__init__(file_path=file_path, sheet_name='studyDesignElements')
      self.items = []
      for index, row in self.sheet.iterrows():
        start_rule = None
        end_rule = None
        xref = self.read_cell_by_name(index, 'xref')
        name = self.read_cell_by_name(index, 'studyElementName')
        description = self.read_cell_by_name(index, 'studyElementDescription')
        start_rule_text = self.read_cell_by_name(index, 'transitionStartRule')
        end_rule_text = self.read_cell_by_name(index, 'transitionEndRule')
        if not start_rule_text == "":
          start_rule = TransitionRule(transitionRuleId=id_manager.build_id(TransitionRule), transitionRuleDescription=start_rule_text)
        if not end_rule_text == "":
          end_rule = TransitionRule(transitionRuleId=id_manager.build_id(TransitionRule), transitionRuleDescription=end_rule_text)
        item = StudyElement(
          studyElementId=id_manager.build_id(StudyElement), 
          studyElementName=name,
          studyElementDescription=description,
          transitionStartRule=start_rule,
          transitionEndRule=end_rule
        )
        self.items.append(item)
        cross_references.add(xref, item)     
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.intercurrent_event import IntercurrentEvent
from usdm_model.analysis_population import AnalysisPopulation
from usdm_model.estimand import Estimand
from usdm_excel.cdisc_ct_library import cdisc_ct_library
from usdm_excel.cdisc_ct import CDISCCT

class StudyDesignEstimandsSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='studyDesignEstimands'))
      super().__init__(file_path=file_path, sheet_name='studyDesignEstimands')
      self.estimands = []
      current = None
      for index, row in self.sheet.iterrows():
        e_summary = self.read_cell_by_name(index, "summaryMeasure")
        ap_description = self.read_cell_by_name(index, "populationDescription")
        ice_name = self.read_cell_by_name(index, "intercurrentEventName")
        ice_description = self.read_cell_by_name(index, "intercurrentEventDescription")
        ice_strategy = self.read_cell_by_name(index, "intercurrentEventStrategy")
        treatment_xref = self.read_cell_by_name(index, "treatmentXref")
        treatment_id = cross_references.get(treatment_xref)
        endpoint_xref = self.read_cell_by_name(index, "endpointXref")
        endpoint_id = cross_references.get(endpoint_xref)
        if not e_summary == "":
          ap = AnalysisPopulation(analysisPopulationId=id_manager.build_id(AnalysisPopulation), populationDescription=ap_description) 
          current = Estimand(estimandId=id_manager.build_id(Estimand), summaryMeasure=e_summary, analysisPopulation=ap, treatment=treatment_id, variableOfInterest=endpoint_id, intercurrentEvents=[])
          self.estimands.append(current)  
        ice = IntercurrentEvent(intercurrentEventId=id_manager.build_id(IntercurrentEvent), intercurrentEventName=ice_name, intercurrentEventDescription=ice_description, intercurrentEventStrategy=ice_strategy)
        current.intercurrentEvents.append(ice)
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.indication import Indication
from usdm_model.investigational_intervention import InvestigationalIntervention

class StudyDesignIISheet(BaseSheet):

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='studyDesignII'))
      super().__init__(file_path=file_path, sheet_name='studyDesignII')
      self.indications = []
      self.interventions = []
      for index, row in self.sheet.iterrows():
        xref = self.read_cell_by_name(index, "xref")
        type = self.read_cell_by_name(index, "type")
        description = self.read_cell_by_name(index, "description")
        #codes = self._build_codes(row, index)
        codes = self.read_other_code_cell_multiple_by_name(index, "codes")
        if type.upper() == "IND":
          item = Indication(indicationId=id_manager.build_id(Indication), indicationDescription=description, codes=codes)
          self.indications.append(item)
          cross_references.add(xref, item.indicationId)
        else:
          item = InvestigationalIntervention(investigationalInterventionId=id_manager.build_id(InvestigationalIntervention), interventionDescription=description, codes=codes)
          self.interventions.append(item)
          cross_references.add(xref, item.investigationalInterventionId)        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




class Error():

  ERROR = 40
  WARNING = 30

  LABEL = { 40: "error", 30: "warning"}

  def __init__(self, sheet: str, row: int, column: int, message: str, level: int=ERROR):
    self.sheet = sheet
    self.row = row
    self.column = column
    self.message = message
    self.level = level

  def to_log(self):
    type = self.__class__.LABEL[self.level].capitalize()  
    if self.sheet == None:
      return f"{type}: {self.message}"
    elif self.row == None:
      return f"{type} in sheet {self.sheet}: {self.message}"
    else:
      return f"{type} in sheet {self.sheet} at [{self.row},{self.column}]: {self.message}"



from typing import Union
from .api_base_model import ApiBaseModel
from .code import Code
from uuid import UUID

class Endpoint(ApiBaseModel):
  endpointId: str
  endpointDescription: str
  endpointPurposeDescription: str
  endpointLevel: Union[Code, None] = None


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.ct_version_manager import ct_version_manager
import traceback
import pandas as pd

class ConfigurationSheet(BaseSheet):

  PARAMS_NAME_COL = 0
  PARAMS_VALUE_COL = 1

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='configuration', header=None))
      super().__init__(file_path=file_path, sheet_name='configuration', header=None)
      self._process_sheet()
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")


  def _process_sheet(self):
    for rindex, row in self.sheet.iterrows():
      name = self.read_cell(rindex, self.PARAMS_NAME_COL)
      value = self.read_cell(rindex, self.PARAMS_VALUE_COL)
      if name.strip().upper() == 'CT VERSION':
        parts = value.split('=')
        if len(parts) == 2:
          ct_version_manager.add(parts[0].strip(), parts[1].strip())
  


from typing import List, Union

from .api_base_model import ApiBaseModel
from .schedule_timeline_exit import ScheduleTimelineExit
from .scheduled_instance import ScheduledInstance

class ScheduleTimeline(ApiBaseModel):
    scheduleTimelineId: str
    scheduleTimelineName: str
    scheduleTimelineDescription: str
    entryCondition: str
    scheduleTimelineEntryId: Union[str, None] = None
    scheduleTimelineExits: List[ScheduleTimelineExit] = []
    scheduleTimelineInstances: List[ScheduledInstance] = []


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_sheet.cycle import Cycle
from usdm_excel.id_manager import id_manager
import pandas as pd

class Cycles():
  
  def __init__(self, parent):
    self.parent = parent
    self.items = []
    timepoint_index = -1
    in_cycle = False
    prev_cycle = None
    for col_index in range(self.parent.sheet.shape[1]):
      if col_index >= SoAColumnRows.FIRST_VISIT_COL:
        timepoint_index += 1
        #cycle, cycle_is_null = self._get_cycle_cell(SoAColumnRows.CYCLE_ROW, col_index)
        cycle, cycle_is_null = self.parent.read_cell_empty_legacy(SoAColumnRows.CYCLE_ROW, col_index)
        if cycle_is_null:
          if in_cycle:
            cycle_record.add_end(self.previous_index(timepoint_index))
            self.items.append(cycle_record)
            in_cycle = False
          else:
            pass # Do nothing
        else:
          cycle = str(cycle)
          if not in_cycle:
            in_cycle = True
            cycle_record = Cycle(self.parent, col_index, cycle, timepoint_index)
          elif prev_cycle == cycle:
            pass # Do nothing
          else:
            cycle_record.add_end(self.previous_index(timepoint_index))
            self.items.append(cycle_record)
            cycle_record = Cycle(self.parent, col_index, cycle, timepoint_index)
        prev_cycle = cycle
      
  # def _get_cycle_cell(self, row_index, col_index):
  #   if self.parent.cell_empty(row_index, col_index):
  #     return "", True
  #   else:
  #     value = self.parent.read_cell_empty(row_index, col_index, '-')
  #     if value == "":
  #       return "", True
  #     else:
  #       return value, False

  def previous_index(self, index):
    if index == 0:
      return 0
    else:
      return index - 1




from typing import Union
from uuid import UUID
from .api_base_model import ApiBaseModel
from .organisation import Organisation

class StudyIdentifier(ApiBaseModel):
  studyIdentifierId: str
  studyIdentifier: str
  studyIdentifierScope: Organisation

  @classmethod
  def scope_reuse(cls):
    return True

  @classmethod
  def search(cls, store, study_uuid):
    results = []
    items = store.get_by_klass_and_scope(cls.__name__, study_uuid)
    #print("IDENTIFIERS:", items)
    for item in items:
      results.append(cls.recursive_read(store, item['uuid']))
    return results




from typing import Union
from .api_base_model import ApiBaseModel
from .transition_rule import TransitionRule

from uuid import UUID

class StudyElement(ApiBaseModel):
  studyElementId: str
  studyElementName: str
  studyElementDescription: str
  transitionStartRule: Union[TransitionRule, None] = None
  transitionEndRule: Union[TransitionRule, None] = None



from .api_base_model import ApiBaseModel
from .code import Code

class Procedure(ApiBaseModel):
  procedureId: str
  procedureType: str
  procedureName: str
  procedureDescription: str
  procedureCode: Code
  procedureIsConditional: bool
  procedureIsConditionalReason: str



from typing import List
from .api_base_model import ApiBaseModel

class BiomedicalConceptSurrogate(ApiBaseModel):
  bcSurrogateId: str
  bcSurrogateName: str
  bcSurrogateDescription: str
  bcSurrogateReference: str



from typing import List, Union
from .api_base_model import ApiBaseModel
from .procedure import Procedure

class Activity(ApiBaseModel):
  activityId: str
  activityName: str
  activityDescription: str
  previousActivityId: Union[str, None] = None
  nextActivityId: Union[str, None] = None
  definedProcedures: List[Procedure] = []
  activityIsConditional: bool
  activityIsConditionalReason: str
  biomedicalConceptIds: List[str] = []
  bcCategoryIds: List[str] = []
  bcSurrogateIds: List[str] = []
  activityTimelineId: Union[str, None] = None


from src.usdm_excel.iso_3166 import ISO3166

def test_code(mocker):
    mock_id = mocker.patch("usdm_excel.id_manager.build_id")
    mock_id.side_effect=['Code_1', 'Code_2', 'Code_3']
    item = ISO3166()
    code = item.code("GB")
    assert code.codeId == "Code_1"
    assert code.code == "GBR"
    assert code.codeSystem == "ISO 3166 1 alpha3"
    assert code.codeSystemVersion == ""
    assert code.decode == "United Kingdom of Great Britain and Northern Ireland"
    code = item.code("GBR")
    assert code.codeId == "Code_2"
    assert code.code == "GBR"
    assert code.codeSystem == "ISO 3166 1 alpha3"
    assert code.codeSystemVersion == ""
    assert code.decode == "United Kingdom of Great Britain and Northern Ireland"
    code = item.code("XXX")
    assert code.codeId == "Code_3"
    assert code.code == "DNK"
    assert code.codeSystem == "ISO 3166 1 alpha3"
    assert code.codeSystemVersion == ""
    assert code.decode == "Denmark"


from typing import List, Union
from .api_base_model import ApiBaseModel
from .study_identifier import *
from .study_protocol_version import *
from .alias_code import *
from .code import Code as genericCode
from .study_design import *
from uuid import UUID

class Study(ApiBaseModel):
  studyId: Union[UUID, None] = None
  studyTitle: str
  studyVersion: str
  studyType: Union[genericCode, None] = None
  studyPhase: Union[AliasCode, None] = None
  businessTherapeuticAreas: List[Code] = []
  studyIdentifiers: List[StudyIdentifier] = []
  studyProtocolVersions: List[StudyProtocolVersion] = []
  studyDesigns: List[StudyDesign] = []
  studyRationale: str
  studyAcronym: str




from usdm_model.organisation import Organisation
from usdm_model.address import Address
from usdm_model.study_identifier import StudyIdentifier
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.iso_3166 import ISO3166
import pandas as pd
import traceback

class StudyIdentifiersSheet(BaseSheet):

  def __init__(self, file_path: str):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='studyIdentifiers'))
      super().__init__(file_path=file_path, sheet_name='studyIdentifiers')
      self.identifiers = []
      self.process_sheet()
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")
      
  def process_sheet(self):
    self.identifiers = []
    for index, row in self.sheet.iterrows():
      organisation_type = self.read_cdisc_klass_attribute_cell_by_name('Organisation', 'organisationType', index, 'organisationType')
      raw_address=self.read_cell_by_name(index, 'organisationAddress')
      organisation = Organisation(
        organisationId=id_manager.build_id(Organisation),
        organisationIdentifierScheme=self.read_cell_by_name(index, 'organisationIdentifierScheme'), 
        organisationIdentifier=self.read_cell_by_name(index, 'organisationIdentifier'),
        organisationName=self.read_cell_by_name(index, 'organisationName'),
        organisationType=organisation_type,
        organizationLegalAddress=self._build_address(raw_address)
      )
      self.identifiers.append(StudyIdentifier(
        studyIdentifierId=id_manager.build_id(StudyIdentifier),
        studyIdentifier=self.read_cell_by_name(index, 'studyIdentifier'), 
        studyIdentifierScope=organisation)
      )
    
  def _build_address(self, raw_address):
    parts = raw_address.split("|")
    if len(parts) == 6:
      return self._to_address(
            id_manager.build_id(Address),
            parts[0].strip(), 
            parts[1].strip(), 
            parts[2].strip(), 
            parts[3].strip(), 
            parts[4].strip(), 
            ISO3166().code(parts[5].strip())
          )
    else:
      return None

  def _to_address(self, id, line, city, district, state, postal_code, country):
    text = "%s, %s, %s, %s, %s, %s" % (line, city, district, state, postal_code, country.decode)
    text = text.replace(' ,', '')
    return Address(addressId=id, text=text, line=line, city=city, district=district, state=state, postalCode=postal_code, country=country)


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_model.activity import Activity as USDMActivity
from usdm_model.biomedical_concept_surrogate import BiomedicalConceptSurrogate
from usdm_excel.cdisc_biomedical_concept import cdisc_bc_library
import pandas as pd

class Activity():
  
  def __init__(self, parent, row_index):
    #super().__init__(sheet)
    #self._row_index = row_index
    self.parent = parent
    self.usdm_biomedical_concept_surrogates = []
    self.usdm_biomedical_concepts = []
    self.name = parent.read_cell(row_index, SoAColumnRows.CHILD_ACTIVITY_COL)
    self._bcs, self._prs, self._tls = self._get_observation_cell(row_index, SoAColumnRows.BC_COL)
    self.usdm_activity = self._as_usdm()
    
  def _as_usdm(self):
    surrogate_bc_items = []
    full_bc_items = []
    procedures = []
    #cdisc_bcs = CDISCBiomedicalConcepts()
    for bc in self._bcs:
      if cdisc_bc_library.exists(bc):
        full_bc = cdisc_bc_library.usdm(bc)
        full_bc_items.append(full_bc.biomedicalConceptId)
        self.usdm_biomedical_concepts.append(full_bc)
      else:
        surrogate = self._to_bc_surrogates(bc)
        surrogate_bc_items.append(surrogate.bcSurrogateId)
        self.usdm_biomedical_concept_surrogates.append(surrogate)
    timelineId = ""
    if len(self._tls) > 0:
      timelineId = cross_references.get(self._tls[0])
    for procedure in self._prs:
      procedures.append(cross_references.get(procedure))
    return USDMActivity(
      activityId=id_manager.build_id(Activity),
      activityName=self.name,
      activityDescription=self.name,
      definedProcedures=procedures,
      activityIsConditional=False,
      activityIsConditionalReason="",
      biomedicalConceptIds=full_bc_items,
      bcCategoryIds=[],
      bcSurrogateIds=surrogate_bc_items,
      activityTimelineId=timelineId
    )
  
  def _to_bc_surrogates(self, name):
    return BiomedicalConceptSurrogate(
      bcSurrogateId=id_manager.build_id(BiomedicalConceptSurrogate),
      bcSurrogateName=name,
      bcSurrogateDescription=name,
      bcSurrogateReference=''
    )
  
  def _get_observation_cell(self, row_index, col_index):
    bcs = []
    prs = []
    tls = []
    if not self.parent.cell_empty(row_index, col_index):
      value = self.parent.read_cell(row_index, col_index)
      outer_parts = value.split(',')
      for outer_part in outer_parts:
        parts = outer_part.split(':')
        if parts[0].strip().upper() == "BC":
          bcs.append(parts[1].strip())
        elif parts[0].strip().upper() == "PR":
          prs.append(parts[1].strip())
        elif parts[0].strip().upper() == "TL":
          tls.append(parts[1].strip())
        else:
          pass
    return bcs, prs, tls



from typing import Union
from .api_base_model import ApiBaseModel
from .code import Code

class Timing(ApiBaseModel):
  timingId: str
  timingType: Code
  timingValue: str
  timingRelativeToFrom: Code
  timingWindow: str
  relativeFromScheduledInstanceId: Union[str, None] = None
  relativeToScheduledInstanceId: Union[str, None] = None


import pytest
from src.usdm_excel.cdisc_biomedical_concept import CDISCBiomedicalConcepts

xfail = pytest.mark.xfail

# def test_code(mocker):
#     mock_id = mocker.patch("usdm_excel.id_manager.build_id")
#     mock_id.side_effect=['Code_1']
#     item = NCIt()
#     code = item.code(code="CODE", decode="DECODE")
#     assert code.codeId == "Code_1"
#     assert code.code == "CODE"
#     assert code.codeSystem == "NCI Thesaurus"
#     assert code.codeSystemVersion == ""
#     assert code.decode == "DECODE"

@xfail
def test_create():
  assert 0

@xfail
def test_exists():
  assert 0

@xfail
def test_catalogue():
  assert 0

@xfail
def test_usdm():
  assert 0

@xfail
def test__get_package_metadata():
  assert 0

@xfail
def test__get_package_items(self):
  assert 0
  
def test__url():
  item = CDISCBiomedicalConcepts()
  assert (item._url('something')) == f"{CDISCBiomedicalConcepts.API_ROOT}something"

def test__bc_as_usdm(mocker):
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['C1', 'BC_1', 'A1', 'C2', 'BC_2', 'A2']
  item = CDISCBiomedicalConcepts()
  data = {
    'conceptId': 'X1',
    'shortName': 'Name',
    'synonym': ['shortie'],
    '_links': { 'self': { 'href': 'link' } }
  }
  bc = item._bc_as_usdm(data)
  assert(bc.biomedicalConceptId) == 'BC_1'
  assert(bc.bcName) == 'Name'
  assert(bc.bcSynonyms) == ['shortie']
  assert(bc.bcReference) == 'link'
  assert(bc.bcConceptCode.aliasCodeId) == 'A1'
  assert(bc.bcConceptCode.standardCode.codeId) == 'C1'
  assert(bc.bcConceptCode.standardCode.code) == 'X1'
  assert(bc.bcConceptCode.standardCode.decode) == 'Name'
  assert(bc.bcConceptCode.standardCode.codeSystem) == "NCI Thesaurus"
  assert(bc.bcConceptCode.standardCode.codeSystemVersion) == "" 
  assert(bc.bcConceptCode.standardCodeAliases) == []
  data = {
    'conceptId': 'X2',
    'shortName': 'Name1',
    '_links': { 'self': { 'href': 'link' } }
  }
  bc = item._bc_as_usdm(data)
  assert(bc.biomedicalConceptId) == 'BC_2'
  assert(bc.bcName) == 'Name1'
  assert(bc.bcSynonyms) == []

@xfail
def test__bc_property_as_usdm(self, property, codes):
  assert 0



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.procedure import Procedure

class StudyDesignProcedureSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='studyDesignProcedures'))
      super().__init__(file_path=file_path, sheet_name='studyDesignProcedures')
      self.procedures = []
      for index, row in self.sheet.iterrows():
        xref = self.read_cell_by_name(index, "xref")
        name = self.read_cell_by_name(index, "procedureName")
        description = self.read_cell_by_name(index, "procedureDescription")
        type = self.read_cell_by_name(index, "procedureType")
        code = self.read_other_code_cell_by_name(index, 'procedureCode')
        conditional = self.read_boolean_cell_by_name(index, 'procedureIsConditional')
        reason = self.read_cell_by_name(index, 'procedureIsConditionalReason')
        item = Procedure(procedureId=id_manager.build_id(Procedure),
          procedureName=name,
          procedureDescription=description,
          procedureType=type, 
          procedureCode=code, 
          procedureIsConditional=conditional, 
          procedureIsConditionalReason=reason
        )
        self.procedures.append(item)
        cross_references.add(xref, item)        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")



from usdm_excel.id_manager import id_manager
from usdm_model.alias_code import AliasCode

class Alias():

  def code(self, standard_code, aliases):
    return AliasCode(aliasCodeId=id_manager.build_id(AliasCode), standardCode=standard_code, standardCodeAliases=aliases)


from typing import List
from .alias_code import AliasCode
from .api_base_model import ApiBaseModel
from .response_code import ResponseCode

class BiomedicalConceptProperty(ApiBaseModel):
  bcPropertyId: str
  bcPropertyName: str
  bcPropertyRequired: bool
  bcPropertyEnabled: bool
  bcPropertyDatatype: str
  bcPropertyResponseCodes: List[ResponseCode] = []
  bcPropertyConceptCode: AliasCode



import pytest

from src.usdm_excel.errors.error import Error

def test_create():
  error = Error(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert error.sheet == "My Sheet"
  assert error.row == 1
  assert error.column == 99
  assert error.message == "XXXXX"

def test_to_log():
  error = Error(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert(error.to_log()) == "Error in sheet My Sheet at [1,99]: XXXXX"
  error = Error(sheet="My Sheet", row=None, column=None, message="XXXXX")
  assert(error.to_log()) == "Error in sheet My Sheet: XXXXX"
  error = Error(sheet=None, row=None, column=None, message="XXXXX")
  assert(error.to_log()) == "Error: XXXXX"
  


from .api_base_model import ApiBaseModel

class IntercurrentEvent(ApiBaseModel):
  intercurrentEventId: str
  intercurrentEventName: str
  intercurrentEventDescription: str
  intercurrentEventStrategy: str



import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="usdm",
  version="0.22.0",
  author="D Iberson-Hurst",
  author_email="",
  description="A python package for using the CDISC TransCelerate USDM",
  long_description=long_description,
  long_description_content_type="text/markdown",
  install_requires=['pandas', 'openpyxl', 'pydantic', 'requests', 'stringcase'],
  packages=setuptools.find_packages(where="src"),
  package_dir={"": "src"},
  package_data={"usdm_excel": ["data/*.yaml", "data/*.json"]},
  tests_require=['pytest'],
  classifiers=[
    "Intended Audience :: Developers",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.10",
    "Operating System :: OS Independent"
  ],
)


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.study_design_population import StudyDesignPopulation

class StudyDesignPopulationSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      #super().__init__(pd.read_excel(open(file_path, 'rb'), sheet_name='studyDesignPopulations'))
      super().__init__(file_path=file_path,  sheet_name='studyDesignPopulations')
      self.populations = []
      for index, row in self.sheet.iterrows():
        description = self.read_cell_by_name(index, "populationDescription")
        number = self.read_cell_by_name(index, "plannedNumberOfParticipants")
        min = self.read_cell_by_name(index, "plannedMinimumAgeOfParticipants")
        max = self.read_cell_by_name(index, "plannedMaximumAgeOfParticipants")
        codes = self._build_codes(row, index)
        self.populations.append(
          StudyDesignPopulation(studyDesignPopulationId=id_manager.build_id(StudyDesignPopulation), 
            populationDescription=description, 
            plannedNumberOfParticipants=number,
            plannedMinimumAgeOfParticipants=min,
            plannedMaximumAgeOfParticipants=max,
            codes=codes
          )
        )
        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")


  def _build_codes(self, row, index):
    #result = []
    return [self.read_cdisc_klass_attribute_cell_by_name('StudyDesignPopulation', "plannedSexOfParticipants", index, "plannedSexOfParticipants")]



from datetime import date
from typing import Union
from .api_base_model import ApiBaseModel
from .code import Code
from uuid import UUID

class StudyProtocolVersion(ApiBaseModel):
  studyProtocolVersionId: str
  briefTitle: str
  officialTitle: str
  publicTitle: str
  scientificTitle: str
  protocolVersion: str
  protocolAmendment: Union[str, None] = None
  protocolEffectiveDate: date
  protocolStatus: Code



from typing import List, Union
from .activity import Activity
from .api_base_model import ApiBaseModel
from .alias_code import AliasCode
from .biomedical_concept import BiomedicalConcept
from .biomedical_concept_category import BiomedicalConceptCategory
from .biomedical_concept_surrogate import BiomedicalConceptSurrogate
from .code import Code
from .encounter import Encounter
from .study_cell import StudyCell
from .indication import Indication
from .investigational_intervention import InvestigationalIntervention
from .study_design_population import StudyDesignPopulation
from .objective import Objective
from .schedule_timeline import ScheduleTimeline
from .estimand import Estimand
import pandas as pd

class StudyDesign(ApiBaseModel):
  studyDesignId: str
  studyDesignName: str
  studyDesignDescription: str
  trialIntentTypes: List[Code] = []
  trialType: List[Code] = []
  interventionModel: Code
  studyCells: List[StudyCell] = []
  studyIndications: List[Indication] = []
  studyInvestigationalInterventions: List[InvestigationalIntervention] = []
  studyPopulations: List[StudyDesignPopulation] = []
  studyObjectives: List[Objective] = []
  studyScheduleTimelines: List[ScheduleTimeline] = []
  therapeuticAreas: List[Code] = []
  studyEstimands: List[Estimand] = []
  encounters: List[Encounter] = []
  activities: List[Activity] = []
  studyDesignRationale: str
  studyDesignBlindingScheme: AliasCode = None
  biomedicalConcepts: List[BiomedicalConcept] = []
  bcCategories: List[BiomedicalConceptCategory] = []
  bcSurrogates: List[BiomedicalConceptSurrogate] = []



from .error import Error
from usdm_excel.logger import package_logger

class Errors():

  WARNING = Error.WARNING
  ERROR = Error.ERROR

  def __init__(self):
    self.items = []

  def clear(self):
    self.items = []

  def add(self, sheet: str, row: int, column: int, message: str, level: int=Error.ERROR) -> None:
    error = Error(sheet, row, column, message, level)
    self.items.append(error)      
    package_logger.log(level, error.to_log())

  def count(self) -> int:
    return len(self.items)
  
  def dump(self):
    result = []
    for item in self.items:
      result.append(item.__dict__)
    return result

error_manager = Errors()


from typing import List, Union
from .api_base_model import ApiBaseModel
from .study_arm import StudyArm
from .study_epoch import StudyEpoch
from .study_element import StudyElement
from uuid import UUID

class StudyCell(ApiBaseModel):
  studyCellId: str
  studyArm: Union[StudyArm, None] = None
  studyEpoch: Union[StudyEpoch, None] = None
  studyElements: List[StudyElement] = []
